local pratcam = {}

pratcam.default = "default"
pratcam.default_count_waits = 6

function pratcam.new(fn,rate,fw,fh,w,h,nearest) -- default expression
  math.randomseed(love.timer.getTime())
  local p = {}

  -- Init
  p.lut = {}
  p.imgf = {} --file
  p.imgc = {} --cache
  p.metacache = setmetatable({},{__mode="v"})

  p.default_count = math.random(0,pratcam.default_count_waits)
  p.bored = {}
  p.nearest = nearest

  p.fw,p.fh = fw,fh -- frame width and height
  p.w,p.h = w,h -- raw image width and height
  p.fcx = p.w / p.fw -- frame count X frames
  p.fcy = p.h / p.fh -- frame count Y frames
  p.fc = p.fcx*p.fcy -- frame count

  p.quads = {}
  for y = 1,p.fcy do
    for x = 1,p.fcx do
      table.insert(p.quads,love.graphics.newQuad(
        (x-1)*p.fw, (y-1)*p.fh, -- zero index
        p.fw, p.fh, -- offset for love's 0.5 pixel offset
         p.w, p.h ) ) -- actual image height
    end
  end

  -- Functions
  p.addCam = pratcam.addCam
  p.setCam = pratcam.setCam
  p._setCam = pratcam._setCam
  p.nextCam = pratcam.nextCam
  p._nextCam = pratcam._nextCam
  p.resetCam = pratcam.resetCam

  p.getWidth = pratcam.getWidth
  p.getHeight = pratcam.getHeight

  p.draw = pratcam.draw
  p.update = pratcam.update

  p._lutLookup = pratcam._lutLookup
  p._isNoLoop = pratcam._isNoLoop

  -- Function Inits
  p:addCam(fn,rate,pratcam.default)
  p:resetCam()

  return p
end

function pratcam:getWidth()
  return self.fw
end

function pratcam:getHeight()
  return self.fh
end

function pratcam:resetCam()
  self:_setCam(1)
  self.frame = math.random(1,self.fc)
end

function pratcam:_lutLookup(expression)
  for i,v in pairs(self.lut) do
    if v == expression then return i end
  end
  assert("Error: Expression `"..expression.."` does not exist.")
end

function pratcam:draw(x,y,scale)
  if not self.imgc[self.camc].data then

    if self.metacache[self.camc] then
      self.imgc[self.camc].data = self.metacache[self.camc]
    else
      self.imgc[self.camc].data = love.graphics.newImage(self.imgf[self.camc])
      self.metacache[self.camc] = self.imgc[self.camc].data
    end 

    if self.nearest then
      self.imgc[self.camc].data:setFilter("nearest","nearest")
    end
  end
  love.graphics.draw(
    self.imgc[self.camc].data,
    self.quads[self.frame],
    x,y,0,scale,scale)
end

function pratcam:_isNoLoop()
  return self.imgc[self.camc].noloop
end

function pratcam:update(dt)

  -- GOOD LUCK TRYING TO FIGURE THIS SHITPILE OUT

  if self.camc == 1 and self.camn then
    self:_setCam(self.camn)
  end

  if self.camc == 1 and self.default_count > pratcam.default_count_waits then
    self:_nextCam( self.bored[math.random(1,#self.bored)] )
  end

  self.t = self.t + dt
  if self.imgc[self.camc].rate and self.t > self.imgc[self.camc].rate then
    self.t = self.t - self.imgc[self.camc].rate
    self.frame = self.frame + 1
    if self:_isNoLoop() and self.frame > self.fc then
      self.frame = self.fc
    end
    if self.frame > self.fc or (self:_isNoLoop() and self.camn) then
      if self.camc == 1 then
        self.default_count = self.default_count + 1
      else
        self.default_count = math.random(0,2)
      end
      if self.camn then
        self:_setCam(self.camn)
      else
        self:_setCam(1)
      end
    end
  end

  -- clear old caches
  for icache,cache in pairs(self.imgc) do
    if icache ~= self.camc and
       icache ~= self.camn and
       cache.age + 1 < love.timer.getTime() then
      cache.data = nil
      cache.age = love.timer.getTime()
    end
  end
end

function pratcam:addCam(fn,rate,expression,noloop,bored)
  assert(type(expression)=="string","Error: argument `expression` should be a string.")
  assert(type(fn)=="string","Error: argument `fn` should be a string.")
  assert(love.filesystem.exists(fn),"Error: File `"..fn.."` does not exist.")
  table.insert(self.lut,expression)
  table.insert(self.imgf,#self.lut,fn)
  table.insert(self.imgc,#self.lut,{
    rate=rate,
    noloop=noloop,
    age=love.timer.getTime()
  })
  if bored then
    table.insert(self.bored,#self.lut)
  end
end

function pratcam:setCam(expression)
  self:_setCam(self:_lutLookup(expression))
end

function pratcam:_setCam(id)
  self.camc = id
  self.t = 0
  self.frame = 1
  self.camn = nil
end

function pratcam:nextCam(expression)
  self:_nextCam(self:_lutLookup(expression))
end

function pratcam:_nextCam(id)
  if id ~= self.camc then
    self.camn = id
  end
end

return pratcam
